@@ -21,35 +21,50 @@
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
  */
 
-
 #include "lime.h"
 
 // This file
-static int write_lime_header(struct resource *);
+static ssize_t write_lime_header(struct resource *);
 static ssize_t write_padding(size_t);
 static void write_range(struct resource *);
 static int init(void);
-ssize_t write_vaddr(void *, size_t);
-int setup(void);
-void cleanup(void);
+static ssize_t write_vaddr(void *, size_t);
+static ssize_t write_flush(void);
+static ssize_t try_write(void *, ssize_t);
+static int setup(void);
+static void cleanup(void);
 
 // External
-extern int write_vaddr_tcp(void *, size_t);
+extern ssize_t write_vaddr_tcp(void *, size_t);
 extern int setup_tcp(void);
 extern void cleanup_tcp(void);
 
-extern int write_vaddr_disk(void *, size_t);
-extern int setup_disk(void);
+extern ssize_t write_vaddr_disk(void *, size_t);
+extern int setup_disk(char *, int);
 extern void cleanup_disk(void);
 
 extern int ldigest_init(void);
 extern int ldigest_update(void *, size_t);
 extern int ldigest_final(void);
+extern int ldigest_write_tcp(void);
+extern int ldigest_write_disk(void);
+extern int ldigest_clean(void);
+
+#ifdef LIME_SUPPORTS_DEFLATE
+extern int deflate_begin_stream(void *, size_t);
+extern int deflate_end_stream(void);
+extern ssize_t deflate(const void *, size_t);
+#endif
 
 static char * format = 0;
 static int mode = 0;
 static int method = 0;
-static char * zero_page;
+
+static void * vpage;
+
+#ifdef LIME_SUPPORTS_DEFLATE
+static void *deflate_page_buf;
+#endif
 
 char * path = 0;
 int dio = 0;
@@ -59,6 +74,8 @@ int localhostonly = 0;
 char * digest = 0;
 int compute_digest = 0;
 
+int no_overlap = 0;
+
 extern struct resource iomem_resource;
 
 module_param(path, charp, S_IRUGO);
@@ -72,11 +89,10 @@ long timeout = 1000;
 module_param(timeout, long, S_IRUGO);
 #endif
 
-#define RETRY_IF_INTURRUPTED(f) ({ \
-    ssize_t err; \
-    do { err = f; } while(err == -EAGAIN || err == -EINTR); \
-    err; \
-})
+#ifdef LIME_SUPPORTS_DEFLATE
+int compress = 0;
+module_param(compress, int, S_IRUGO);
+#endif
 
 int init_module (void)
 {
@@ -101,7 +117,9 @@ int init_module (void)
     DBG("  TIMEOUT: %lu", timeout);
 #endif
 
-    zero_page = kzalloc(PAGE_SIZE, GFP_KERNEL);
+#ifdef LIME_SUPPORTS_DEFLATE
+    DBG("  COMPRESS: %u", compress);
+#endif
 
     if (!strcmp(format, "raw")) mode = LIME_MODE_RAW;
     else if (!strcmp(format, "lime")) mode = LIME_MODE_LIME;
@@ -113,6 +131,7 @@ int init_module (void)
 
     method = (sscanf(path, "tcp:%d", &port) == 1) ? LIME_METHOD_TCP : LIME_METHOD_DISK;
     if (digest) compute_digest = LIME_DIGEST_COMPUTE;
+
     return init();
 }
 
@@ -127,26 +146,42 @@ static int init() {
 
     DBG("Initializing Dump...");
 
-    if((err = setup())) {
+    if ((err = setup())) {
         DBG("Setup Error");
         cleanup();
         return err;
     }
 
-    if(compute_digest == LIME_DIGEST_COMPUTE)
+    if (digest) {
         compute_digest = ldigest_init();
+        no_overlap = 1;
+    }
+
+    vpage = (void *) __get_free_page(GFP_NOIO);
+
+#ifdef LIME_SUPPORTS_DEFLATE
+    if (compress) {
+        deflate_page_buf = kmalloc(PAGE_SIZE, GFP_NOIO);
+        err = deflate_begin_stream(deflate_page_buf, PAGE_SIZE);
+        if (err < 0) {
+            DBG("ZLIB begin stream failed");
+            return err;
+        }
+        no_overlap = 1;
+    }
+#endif
 
     for (p = iomem_resource.child; p ; p = p->sibling) {
 
         if (strcmp(p->name, LIME_RAMSTR))
             continue;
 
-        if (mode == LIME_MODE_LIME && (err = write_lime_header(p))) {
+        if (mode == LIME_MODE_LIME && write_lime_header(p) < 0) {
             DBG("Error writing header 0x%lx - 0x%lx", (long) p->start, (long) p->end);
-           break;
-        } else if (mode == LIME_MODE_PADDED && (err = write_padding((size_t) ((p->start - 1) - p_last)))) {
+            break;
+        } else if (mode == LIME_MODE_PADDED && write_padding((size_t) ((p->start - 1) - p_last)) < 0) {
             DBG("Error writing padding 0x%lx - 0x%lx", (long) p_last, (long) p->start - 1);
-           break;
+            break;
         }
 
         write_range(p);
@@ -154,19 +189,43 @@ static int init() {
         p_last = p->end;
     }
 
+    write_flush();
+
     DBG("Memory Dump Complete...");
 
     cleanup();
 
-    if(compute_digest == LIME_DIGEST_COMPUTE)
+    if (compute_digest == LIME_DIGEST_COMPUTE) {
+        DBG("Writing Out Digest.");
+
         compute_digest = ldigest_final();
 
-    return err;
+        if (compute_digest == LIME_DIGEST_COMPLETE) {
+            if (method == LIME_METHOD_TCP)
+                err = ldigest_write_tcp();
+            else
+                err = ldigest_write_disk();
+
+            DBG("Digest Write %s.", (err == 0) ? "Complete" : "Failed");
+        }
+    }
+
+    if (digest)
+        ldigest_clean();
+
+#ifdef LIME_SUPPORTS_DEFLATE
+    if (compress) {
+        deflate_end_stream();
+        kfree(deflate_page_buf);
+    }
+#endif
+
+    free_page((unsigned long) vpage);
+
+    return 0;
 }
 
-static int write_lime_header(struct resource * res) {
-    ssize_t s;
-
+static ssize_t write_lime_header(struct resource * res) {
     lime_mem_range_header header;
 
     memset(&header, 0, sizeof(lime_mem_range_header));
@@ -175,24 +234,19 @@ static int write_lime_header(struct resource * res) {
     header.s_addr = res->start;
     header.e_addr = res->end;
 
-    s = write_vaddr(&header, sizeof(lime_mem_range_header));
-
-    if (s != sizeof(lime_mem_range_header)) {
-        DBG("Error sending header %zd", s);
-        return (int) s;
-    }
-
-    return 0;
+    return write_vaddr(&header, sizeof(lime_mem_range_header));
 }
 
 static ssize_t write_padding(size_t s) {
     size_t i = 0;
     ssize_t r;
 
+    memset(vpage, 0, PAGE_SIZE);
+
     while(s -= i) {
 
         i = min((size_t) PAGE_SIZE, s);
-        r = write_vaddr(zero_page, i);
+        r = write_vaddr(vpage, i);
 
         if (r != i) {
             DBG("Error sending zero page: %zd", r);
@@ -235,25 +289,21 @@ static void write_range(struct resource * res) {
             write_padding(is);
         } else {
             v = kmap(p);
-            //If we don't need to compute the digest; lets save some memory 
-            //and cycles
-            if(compute_digest == LIME_DIGEST_COMPUTE) {
-                void * lv = kmalloc(is, GFP_ATOMIC);
-                memcpy(lv, v, is);
-                s = write_vaddr(lv, is);
-                kfree(lv);
+            /*
+             * If we need to compute the digest or compress the output
+             * take a snapshot of the page. Otherwise save some cycles.
+             */
+            if (no_overlap) {
+                copy_page(vpage, v);
+                s = write_vaddr(vpage, is);
             } else {
                 s = write_vaddr(v, is);
             }
-
-            kunmap(p);            
+            kunmap(p);
 
             if (s < 0) {
-                DBG("Error writing page: vaddr %p ret: %zd.  Null padding.", v, s);
-                write_padding(is);
-            } else if (s != is) {
-                DBG("Short Read %zu instead of %lu.  Null padding.", s, (unsigned long) is);
-                write_padding(is - s);
+                DBG("Failed to write page: vaddr %p. Skipping Range...", v);
+                break;
             }
         }
 
@@ -270,20 +320,62 @@ static void write_range(struct resource * res) {
     }
 }
 
-ssize_t write_vaddr(void * v, size_t is) {
-    if(compute_digest == LIME_DIGEST_COMPUTE)
+static ssize_t write_vaddr(void * v, size_t is) {
+    ssize_t ret;
+
+    if (compute_digest == LIME_DIGEST_COMPUTE)
         compute_digest = ldigest_update(v, is);
 
-    return RETRY_IF_INTURRUPTED(
+#ifdef LIME_SUPPORTS_DEFLATE
+    if (compress) {
+        /* Run deflate() on input until output buffer is not full. */
+        do {
+            ret = try_write(deflate_page_buf, deflate(v, is));
+            if (ret < 0)
+                return ret;
+        } while (ret == PAGE_SIZE);
+        return is;
+    }
+#endif
+
+    ret = try_write(v, is);
+    return ret;
+}
+
+static ssize_t write_flush(void) {
+#ifdef LIME_SUPPORTS_DEFLATE
+    if (compress) {
+        try_write(deflate_page_buf, deflate(NULL, 0));
+    }
+#endif
+    return 0;
+}
+
+static ssize_t try_write(void * v, ssize_t is) {
+    ssize_t ret;
+
+    if (is <= 0)
+        return is;
+
+    ret = RETRY_IF_INTERRUPTED(
         (method == LIME_METHOD_TCP) ? write_vaddr_tcp(v, is) : write_vaddr_disk(v, is)
     );
+
+    if (ret < 0) {
+        DBG("Write error: %zd", ret);
+    } else if (ret != is) {
+        DBG("Short write %zu instead of %zu.", ret, is);
+        ret = -1;
+    }
+
+    return ret;
 }
 
-int setup(void) {
-    return (method == LIME_METHOD_TCP) ? setup_tcp() : setup_disk();
+static int setup(void) {
+    return (method == LIME_METHOD_TCP) ? setup_tcp() : setup_disk(path, dio);
 }
 
-void cleanup(void) {
+static void cleanup(void) {
     return (method == LIME_METHOD_TCP) ? cleanup_tcp() : cleanup_disk();
 }
 
